import yaml
from pprint import pprint
import pandas as pd
import torch
import torchio

from GANDLF.parseConfig import parseConfig
from GANDLF.utils import find_problem_type_from_parameters, one_hot
from GANDLF.metrics import overall_stats
from GANDLF.losses.segmentation import dice
from GANDLF.metrics.segmentation import (
    _calculator_generic_all_surface_distances,
    _calculator_sensitivity_specificity,
    _calculator_jaccard,
)


def generate_metrics_dict(input_csv: str, config: str, outputfile: str = None) -> dict:
    """
    This function generates metrics from the input csv and the config.

    Args:
        input_csv (str): The input CSV.
        config (str): The input yaml config.
        outputfile (str, optional): The output file to save the metrics. Defaults to None.

    Returns:
        dict: The metrics dictionary.
    """
    input_df = pd.read_csv(input_csv)

    # check required headers in a case insensitive manner
    headers = {}
    required_columns = ["subjectid", "prediction", "target"]
    for col, _ in input_df.iteritems():
        col_lower = col.lower()
        for column_to_check in required_columns:
            if column_to_check == col_lower:
                headers[column_to_check] = col
    for column in required_columns:
        assert column in headers, f"The input csv should have a column named {column}"

    overall_stats_dict = {}
    parameters = parseConfig(config)
    problem_type = find_problem_type_from_parameters(parameters)
    parameters["problem_type"] = problem_type

    if problem_type == "regression" or problem_type == "classification":
        parameters["model"]["num_classes"] = len(parameters["model"]["class_list"])
        predictions_tensor = torch.from_numpy(
            input_df[headers["prediction"]].to_numpy().ravel()
        )
        labels_tensor = torch.from_numpy(input_df[headers["target"]].to_numpy().ravel())
        overall_stats_dict = overall_stats(
            predictions_tensor, labels_tensor, parameters
        )
    elif problem_type == "segmentation":
        # read images and then calculate metrics
        class_list = parameters["model"]["class_list"]
        for _, row in input_df.iterrows():
            current_subject_id = row[headers["subjectid"]]
            overall_stats_dict[current_subject_id] = {}
            label_image = torchio.LabelMap(row["target"])
            pred_image = torchio.LabelMap(row["prediction"])
            label_tensor = label_image.data
            pred_tensor = pred_image.data
            spacing = label_image.spacing
            if label_tensor.data.shape[-1] == 1:
                spacing = spacing[0:2]
            # add dimension for batch
            parameters["subject_spacing"] = torch.Tensor(spacing).unsqueeze(0)
            label_tensor = label_tensor.unsqueeze(0)
            pred_tensor = pred_tensor.unsqueeze(0)

            # one hot encode with batch_size = 1
            label_image_one_hot = one_hot(label_tensor, class_list)
            pred_image_one_hot = one_hot(pred_tensor, class_list)

            for class_index, _ in enumerate(class_list):
                overall_stats_dict[current_subject_id][str(class_index)] = {}
                # this is inconsequential, since one_hot will ensure that the classes are present
                parameters["model"]["class_list"] = [0, 1]
                parameters["model"]["num_classes"] = 2
                parameters["model"]["ignore_label_validation"] = 0
                overall_stats_dict[current_subject_id][str(class_index)]["dice"] = dice(
                    pred_image_one_hot,
                    label_image_one_hot,
                ).item()
                nsd, hd100, hd95 = _calculator_generic_all_surface_distances(
                    pred_image_one_hot,
                    label_image_one_hot,
                    parameters,
                )
                overall_stats_dict[current_subject_id][str(class_index)][
                    "nsd"
                ] = nsd.item()
                overall_stats_dict[current_subject_id][str(class_index)][
                    "hd100"
                ] = hd100.item()
                overall_stats_dict[current_subject_id][str(class_index)][
                    "hd95"
                ] = hd95.item()

                (
                    s,
                    p,
                ) = _calculator_sensitivity_specificity(
                    pred_image_one_hot,
                    label_image_one_hot,
                    parameters,
                )
                overall_stats_dict[current_subject_id][str(class_index)][
                    "sensitivity"
                ] = s.item()
                overall_stats_dict[current_subject_id][str(class_index)][
                    "specificity"
                ] = p.item()
                overall_stats_dict[current_subject_id][
                    "jaccard_" + str(class_index)
                ] = _calculator_jaccard(
                    pred_image_one_hot,
                    label_image_one_hot,
                    parameters,
                ).item()

    pprint(overall_stats_dict)
    if outputfile is not None:
        with open(outputfile, "w") as outfile:
            yaml.dump(overall_stats_dict, outfile)



from torch.optim import (
    SGD,
    ASGD,
    Rprop,
    Adam,
    AdamW,
    # SparseAdam,
    Adamax,
    Adadelta,
    Adagrad,
    RMSprop,
    RAdam,
)


def sgd(parameters):
    """
    Creates a Stochastic Gradient Descent optimizer from the PyTorch `torch.optim` module using the input parameters.

    Args:
        parameters (dict): A dictionary containing the input parameters for the optimizer.

    Returns:
        optimizer (torch.optim.SGD): A Stochastic Gradient Descent optimizer.

    """
    # Create the optimizer using the input parameters
    optimizer = SGD(
        parameters["model_parameters"],
        lr=parameters.get("learning_rate"),
        momentum=parameters["optimizer"].get("momentum", 0.9),
        weight_decay=parameters["optimizer"].get("weight_decay", 0),
        dampening=parameters["optimizer"].get("dampening", 0),
        nesterov=parameters["optimizer"].get("Nesterov", False),
    )

    return optimizer


def asgd(parameters):
    """
    Creates an Averaged Stochastic Gradient Descent optimizer from the PyTorch `torch.optim` module using the input parameters.

    Args:
        parameters (dict): A dictionary containing the input parameters for the optimizer.

    Returns:
        optimizer (torch.optim.ASGD): An Averaged Stochastic Gradient Descent optimizer.

    """
    # Create the optimizer using the input parameters
    return ASGD(
        parameters["model_parameters"],
        lr=parameters.get("learning_rate"),
        alpha=parameters["optimizer"].get("alpha", 0.75),
        t0=parameters["optimizer"].get("t0", 1e6),
        lambd=parameters["optimizer"].get("lambd", 1e-4),
        weight_decay=parameters["optimizer"].get("weight_decay", 0),
    )


def adam(parameters, opt_type="normal"):
    """
    Creates an Adam or AdamW optimizer from the PyTorch `torch.optim` module using the input parameters.

    Args:
        parameters (dict): A dictionary containing the input parameters for the optimizer.
        opt_type (str): A string indicating the type of optimizer to create (either "normal" for Adam or "AdamW" for AdamW).

    Returns:
        optimizer (torch.optim.Adam or torch.optim.AdamW): An Adam or AdamW optimizer.

    """
    # Determine which optimizer to create based on opt_type
    if opt_type == "normal":
        optimizer_fn = Adam
    elif opt_type == "AdamW":
        optimizer_fn = AdamW
    else:
        raise ValueError(f"Invalid optimizer type: {opt_type}")

    # Create the optimizer using the input parameters
    return optimizer_fn(
        parameters["model_parameters"],
        lr=parameters.get("learning_rate"),
        betas=parameters["optimizer"].get("betas", (0.9, 0.999)),
        weight_decay=parameters["optimizer"].get("weight_decay", 0.00005),
        eps=parameters["optimizer"].get("eps", 1e-8),
        amsgrad=parameters["optimizer"].get("amsgrad", False),
    )


def adamw(parameters):
    """
    Creates an AdamW optimizer from the PyTorch `torch.optim` module using the input parameters.

    Args:
        parameters (dict): A dictionary containing the input parameters for the optimizer.

    Returns:
        optimizer (torch.optim.AdamW): An AdamW optimizer.

    """
    return adam(parameters, opt_type="AdamW")


def adamax(parameters):
    """
    Creates an Adamax optimizer from the PyTorch `torch.optim` module using the input parameters.

    Args:
        parameters (dict): A dictionary containing the input parameters for the optimizer.

    Returns:
        optimizer (torch.optim.Adamax): An Adamax optimizer.

    """
    # Create the optimizer using the input parameters
    return Adamax(
        parameters["model_parameters"],
        lr=parameters.get("learning_rate"),
        betas=parameters["optimizer"].get("betas", (0.9, 0.999)),
        weight_decay=parameters["optimizer"].get("weight_decay", 0.00005),
        eps=parameters["optimizer"].get("eps", 1e-8),
    )


# def sparseadam(parameters):
#     # pick defaults
#     if not ("betas" in parameters["optimizer"]):
#         parameters["optimizer"]["betas"] = (0.9, 0.999)
#     if not ("eps" in parameters["optimizer"]):
#         parameters["optimizer"]["eps"] = 1e-8

#     return SparseAdam(
#         parameters["model_parameters"],
#         lr=parameters["learning_rate"],
#         betas=parameters["optimizer"]["betas"],
#         eps=parameters["optimizer"]["eps"],
#     )


def rprop(parameters):
    """
    Creates a Resilient Backpropagation optimizer from the PyTorch `torch.optim` module using the input parameters.

    Args:
        parameters (dict): A dictionary containing the input parameters for the optimizer.

    Returns:
        optimizer (torch.optim.Rprop): A Resilient Backpropagation optimizer.

    """
    # Create the optimizer using the input parameters
    return Rprop(
        parameters["model_parameters"],
        lr=parameters.get("learning_rate"),
        etas=parameters["optimizer"].get("etas", (0.5, 1.2)),
        step_sizes=parameters["optimizer"].get("step_sizes", (1e-7, 50)),
    )


def adadelta(parameters):
    """
    Creates an Adadelta optimizer from the PyTorch `torch.optim` module using the input parameters.

    Args:
        parameters (dict): A dictionary containing the input parameters for the optimizer.

    Returns:
        optimizer (torch.optim.Adadelta): An Adadelta optimizer.

    """
    # Create the optimizer using the input parameters
    return Adadelta(
        parameters["model_parameters"],
        lr=parameters.get("learning_rate"),
        rho=parameters["optimizer"].get("rho", 0.9),
        eps=parameters["optimizer"].get("eps", 1e-6),
        weight_decay=parameters["optimizer"].get("weight_decay", 0),
    )


def adagrad(parameters):
    """
    Creates an Adagrad optimizer from the PyTorch `torch.optim` module using the input parameters.

    Args:
        parameters (dict): A dictionary containing the input parameters for the optimizer.

    Returns:
        optimizer (torch.optim.Adagrad): An Adagrad optimizer.

    """

    # Create the optimizer using the input parameters
    return Adagrad(
        parameters["model_parameters"],
        lr=parameters.get("learning_rate"),
        lr_decay=parameters["optimizer"].get("lr_decay", 0),
        eps=parameters["optimizer"].get("eps", 1e-6),
        weight_decay=parameters["optimizer"].get("weight_decay", 0),
    )


def rmsprop(parameters):
    """
    Creates an RMSprop optimizer from the PyTorch `torch.optim` module using the input parameters.

    Args:
        parameters (dict): A dictionary containing the input parameters for the optimizer.

    Returns:
        optimizer (torch.optim.RMSprop): An RMSprop optimizer.

    """
    # Create the optimizer using the input parameters
    return RMSprop(
        parameters["model_parameters"],
        lr=parameters.get("learning_rate"),
        alpha=parameters["optimizer"].get("alpha", 0.99),
        eps=parameters["optimizer"].get("eps", 1e-8),
        centered=parameters["optimizer"].get("centered", False),
        momentum=parameters["optimizer"].get("momentum", 0),
        weight_decay=parameters["optimizer"].get("weight_decay", 0),
    )


def radam(parameters):
    """
    Creates a RAdam optimizer from the PyTorch `torch.optim` module using the input parameters.

    Args:
        parameters (dict): A dictionary containing the input parameters for the optimizer.

    Returns:
        optimizer (torch.optim.RAdam): A RAdam optimizer.
    """
    # Create the optimizer using the input parameters
    return RAdam(
        parameters["model_parameters"],
        lr=parameters.get("learning_rate"),
        betas=parameters["optimizer"].get("betas", (0.9, 0.999)),
        eps=parameters["optimizer"].get("eps", 1e-8),
        weight_decay=parameters["optimizer"].get("weight_decay", 0),
        foreach=parameters["optimizer"].get("foreach", None),
    )
